# -*- coding: utf-8 -*-
"""
/*------------------------------------------------------*
|                         BRAILS                        |
|                                                       |
| Author: Charles Wang,  UC Berkeley, c_w@berkeley.edu  |
|                                                       |
| Date:    1/19/2021                                   |
*------------------------------------------------------*/
"""


stateAbbrs =  {'Alabama': 'AL',
     'Alaska': 'AK',
     'Arizona': 'AZ',
     'Arkansas': 'AR',
     'California': 'CA',
     'Colorado': 'CO',
     'Connecticut': 'CT',
     'Delaware': 'DE',
     'District of Columbia': 'DC',
     'Florida': 'FL',
     'Georgia': 'GA',
     'Hawaii': 'HI',
     'Idaho': 'ID',
     'Illinois': 'IL',
     'Indiana': 'IN',
     'Iowa': 'IA',
     'Kansas': 'KS',
     'Kentucky': 'KY',
     'Louisiana': 'LA',
     'Maine': 'ME',
     'Maryland': 'MD',
     'Massachusetts': 'MA',
     'Michigan': 'MI',
     'Minnesota': 'MN',
     'Mississippi': 'MS',
     'Missouri': 'MO',
     'Montana': 'MT',
     'Nebraska': 'NE',
     'Nevada': 'NV',
     'New Hampshire': 'NH',
     'New Jersey': 'NJ',
     'New Mexico': 'NM',
     'New York': 'NY',
     'North Carolina': 'NC',
     'North Dakota': 'ND',
     'Ohio': 'OH',
     'Oklahoma': 'OK',
     'Oregon': 'OR',
     'Pennsylvania': 'PA',
     'Rhode Island': 'RI',
     'South Carolina': 'SC',
     'South Dakota': 'SD',
     'Tennessee': 'TN',
     'Texas': 'TX',
     'Utah': 'UT',
     'Vermont': 'VT',
     'Virginia': 'VA',
     'Washington': 'WA',
     'West Virginia': 'WV',
     'Wisconsin': 'WI',
     'Wyoming': 'WY'}

stateNames = {'AL': 'Alabama',
     'AK': 'Alaska',
     'AZ': 'Arizona',
     'AR': 'Arkansas',
     'CA': 'California',
     'CO': 'Colorado',
     'CT': 'Connecticut',
     'DE': 'Delaware',
     'DC': 'District of Columbia',
     'FL': 'Florida',
     'GA': 'Georgia',
     'HI': 'Hawaii',
     'ID': 'Idaho',
     'IL': 'Illinois',
     'IN': 'Indiana',
     'IA': 'Iowa',
     'KS': 'Kansas',
     'KY': 'Kentucky',
     'LA': 'Louisiana',
     'ME': 'Maine',
     'MD': 'Maryland',
     'MA': 'Massachusetts',
     'MI': 'Michigan',
     'MN': 'Minnesota',
     'MS': 'Mississippi',
     'MO': 'Missouri',
     'MT': 'Montana',
     'NE': 'Nebraska',
     'NV': 'Nevada',
     'NH': 'New Hampshire',
     'NJ': 'New Jersey',
     'NM': 'New Mexico',
     'NY': 'New York',
     'NC': 'North Carolina',
     'ND': 'North Dakota',
     'OH': 'Ohio',
     'OK': 'Oklahoma',
     'OR': 'Oregon',
     'PA': 'Pennsylvania',
     'RI': 'Rhode Island',
     'SC': 'South Carolina',
     'SD': 'South Dakota',
     'TN': 'Tennessee',
     'TX': 'Texas',
     'UT': 'Utah',
     'VT': 'Vermont',
     'VA': 'Virginia',
     'WA': 'Washington',
     'WV': 'West Virginia',
     'WI': 'Wisconsin',
     'WY': 'Wyoming'}

footprintFiles = {'Alabama': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Alabama.zip',
 'Alaska': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Alaska.zip',
 'Arizona': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Arizona.zip',
 'Arkansas': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Arkansas.zip',
 'California': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/California.zip',
 'Colorado': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Colorado.zip',
 'Connecticut': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Connecticut.zip',
 'Delaware': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Delaware.zip',
 'District Of Columbia': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/DistrictofColumbia.zip',
 'Florida': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Florida.zip',
 'Georgia': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Georgia.zip',
 'Hawaii': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Hawaii.zip',
 'Idaho': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Idaho.zip',
 'Illinois': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Illinois.zip',
 'Indiana': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Indiana.zip',
 'Iowa': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Iowa.zip',
 'Kansas': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Kansas.zip',
 'Kentucky': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Kentucky.zip',
 'Louisiana': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Louisiana.zip',
 'Maine': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Maine.zip',
 'Maryland': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Maryland.zip',
 'Massachusetts': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Massachusetts.zip',
 'Michigan': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Michigan.zip',
 'Minnesota': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Minnesota.zip',
 'Mississippi': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Mississippi.zip',
 'Missouri': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Missouri.zip',
 'Montana': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Montana.zip',
 'Nebraska': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Nebraska.zip',
 'Nevada': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Nevada.zip',
 'New Hampshire': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/NewHampshire.zip',
 'New Jersey': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/NewJersey.zip',
 'New Mexico': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/NewMexico.zip',
 'New York': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/NewYork.zip',
 'North Carolina': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/NorthCarolina.zip',
 'North Dakota': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/NorthDakota.zip',
 'Ohio': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Ohio.zip',
 'Oklahoma': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Oklahoma.zip',
 'Oregon': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Oregon.zip',
 'Pennsylvania': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Pennsylvania.zip',
 'Rhode Island': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/RhodeIsland.zip',
 'South Carolina': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/SouthCarolina.zip',
 'South Dakota': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/SouthDakota.zip',
 'Tennessee': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Tennessee.zip',
 'Texas': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Texas.zip',
 'Utah': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Utah.zip',
 'Vermont': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Vermont.zip',
 'Virginia': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Virginia.zip',
 'Washington': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Washington.zip',
 'West Virginia': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/WestVirginia.zip',
 'Wisconsin': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Wisconsin.zip',
 'Wyoming': 'https://usbuildingdata.blob.core.windows.net/usbuildings-v1-1/Wyoming.zip'}