from .folder import ImageFolderInstance, MaskFolderInstance, CombinedMaskDataset
from .cifar import CIFAR10Instance, CIFAR100Instance
from .mnist import MNISTInstance

__all__ = ('ImageFolderInstance', 'MaskFolderInstance', 'CombinedMaskDataset', 'MNISTInstance', 'CIFAR10Instance', 'CIFAR100Instance')

